/* a simple chat app using allnet */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/select.h>

#include "packet.h"
#include "achat.h"
#include "api.h"
#include "app_storage.h"

/* maximum string size for a variety of things */
#define MAGIC_SIZE	141

static char prompt [MAGIC_SIZE];
static int waiting_for_ack = 0;
static char sent_message_owner_id [PACKET_ID_SIZE];
static char expecting_peer [MAGIC_SIZE];
static char expecting_secret [MAGIC_SIZE];
static char peer_shared_secret [MAGIC_SIZE];
static int exchanging_key_id = -1;
static int received_key_count = -1;

#if 0
/* while initializing don't print, print messages later in reverse order */
static int initializing = 1;
/* when initializing, save the earliers counter value received from the past */
static message_counter init_received_counter = 0;
#define MAX_SAVED_MESSAGES	1000
struct message_info {
  char string [MAGIC_SIZE];
  message_time gtime;
  message_counter counter;
  int key_id;
};
static struct message_info messages [MAX_SAVED_MESSAGES];
static int saved_messages = 0;
#endif /* 0 */

static pthread_cond_t ack_cond = PTHREAD_COND_INITIALIZER;

static void strip_non_alpha (char * s)
{
  char * original = s;
  while (*s != '\0') {
    if (! isalpha (*s)) {
      char * from = s + 1;
      char * to = s;
      while (*from != '\0')
        *(to++) = *(from++);
      *to = '\0';
    } else {
      s++;
    }
  }
}

static void make_upper (char * s)
{
  while (*s != '\0') {
    *s = toupper (*s);
    s++;
  }
}

static void make_lower (char * s)
{
  while (*s != '\0') {
    *s = tolower (*s);
    s++;
  }
}

static int local_time_offset (struct tm * ltime_tm, time_t * gtime)
{
  struct tm gtime_tm;
  gmtime_r (gtime, &gtime_tm);
  time_t ltime = mktime (ltime_tm);
  time_t gtime2 = mktime (&gtime_tm);
/*
  printf ("gtime is %ld = %s, gtime 2 is %ld\n",
          *gtime, asctime (&gtime_tm), gtime2);
  printf ("ltime is %ld = %d:%02d = %s\n", ltime,
          ltime_tm->tm_hour, ltime_tm->tm_min, asctime (ltime_tm));
*/
  /* printf ("local time offset is %ld\n", (ltime - gtime2) / 60); */
  return (ltime - gtime2) / 60;
}

static void send_key_hash (char * shared_secret, int key_id, int hops)
{
  char destination [DESTINATION_SIZE];
  bzero (destination, DESTINATION_SIZE);
  allnet_send_key (destination, 0, hops, shared_secret, key_id, NULL, 0);
  /*
  *** to be done
  questions:  why is there a nonce?  and should it be a timestamp,
              or the contact name?  should it be displayed to the user?
  */
}

/* in case of error, returns NULL.  Otherwise, returnes a malloc'd string
 * with the user input.  Never returns an empty string, that is, a string
 * with the first character set to '\0'.
 *
 * if finished is NULL, only returns after it gets a line.
 * if finished is not NULL, returns either after it gets a line, or
 * when *finished is -1, whichever happens first.  In the latter case,
 * returns NULL.
 *
 * if finished and message are not NULL, prints message if it returns
 * because *finished was set to -1
 */
static char * my_readline (char * prompt, int * finished, char * message)
{
  int alloc_size = 5;
  char * result = malloc (alloc_size);
  *result = '\0';
  printf ("%s", prompt);
  char * p = result;
  int nc = 0;
  while (1) {
    fd_set read_fds;
    FD_ZERO (&read_fds);
    FD_SET (STDIN_FILENO, &read_fds);
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10000;  /* 10 ms, for fast response */
    struct timeval *tp = &t;
    if (finished == NULL)   /* nothing to time out for */
      tp = NULL;
    if (select (STDIN_FILENO + 1, &read_fds, NULL, NULL, tp) == 0) {
       /* timeout, check to see if we are finished */
      if ((finished != NULL) && (*finished == -1)) {
        if (message != NULL)
          printf ("\n%s\n", message);
        free (result);
        return NULL;
      }
      continue;  /* do not read, restart executing the while loop */
    }
    if (read (STDIN_FILENO, p, 1) != 1) {
      perror ("read/stdin");
      free (result);
      return NULL;
    }
    if (*p == '\n') {
      *p = '\0';
      if (p == result) {
        free (result);
        return NULL;
      }  /* else result is not empty */
      /* printf ("my_readline returning '%s'\n", result); */
      return result;
    }
    p++;
    *p = '\0';
    nc++;
    if (nc + 1 > alloc_size) {
      alloc_size = alloc_size * 2 + 5;
      result = realloc (result, alloc_size);
      p = result + nc;
    }
  }
}

static void save_message (char * string, message_time gtime, int counter,
                          int key)
{
  /* save the counter, if it is earlier than anything we have seen */
  if ((init_received_counter == 0) ||
      (counter < init_received_counter))
    init_received_counter = counter;

  int i;
  for (i = 0; i < saved_messages; i++) {
    if ((key == messages [i].key_id) &&
        (gtime == messages [i].gtime) &&
        (counter == messages [i].counter) &&
        (strcmp (string, messages [i].string) == 0)) {
      printf ("message already saved\n");
      return;
    }
  }
  if (saved_messages < MAX_SAVED_MESSAGES) {
    strncpy_t (messages [saved_messages].string, string, MAGIC_SIZE);
    messages [i].gtime = gtime;
    messages [i].counter = counter;
    messages [i].key_id = key;
    saved_messages++;
  } else {
    printf ("too many messages, unable to save %s\n", string);
  }
}

static void msg_to_string (int key_id, message_counter sent_c,
                           message_time sent_time, char * msg,
                           char * result, int rsize)
{
  char time_string [MAGIC_SIZE];
  time_to_string (time_string, MAGIC_SIZE, sent_time);

  char name [MAGIC_SIZE] = "unknown";
  char user_path [MAGIC_SIZE] = "";
  name_for_key (key_id, name, MAGIC_SIZE - 1, user_path, MAGIC_SIZE - 1);

  snprintf (result, rsize,
            "(%lld) from %s at %s: %s", sent_c, name, time_string, msg);
}

static void handle_user_message (char * buffer, int bsize,
                                 struct allnet_data_header * hp)
{
/*
  printf ("received user message: ");
  print_buffer (buffer, bsize, 1);
*/
  if (bsize <= ALLNET_DATA_HEADER_SIZE + ACHAT_HEADER_SIZE) {
    printf ("useless chat message of size %d, header size %d+%d = %d\n",
            bsize, ALLNET_DATA_HEADER_SIZE, ACHAT_HEADER_SIZE,
            ALLNET_DATA_HEADER_SIZE + ACHAT_HEADER_SIZE);
    return;
  }
  int cipher = buffer + ALLNET_DATA_HEADER_SIZE;
  int csize = bsize - ALLNET_DATA_HEADER_SIZE;
  int tsize = dsize;
  char * text_buffer = malloc (tsize);
  if (text_buffer == NULL) {
    printf ("unable to allocate %d bytes for decrypt text buffer\n", tsize);
    return;
  }
  int rsize;  /* size of the decrypted data in text_buffer */
  int key_id; /* matching key ID */
  decrypt_verify (cipher, csize, text_buffer, tsize,
                  hp->destination, hp->num_destination_bits, &rsize, &key_id);
  if (rsize == -1) {
    /* too many to print, including my own messages sent back to me
     printf ("unable to decrypt incoming message, discarding\n"); */
    return;
  }
  if (rsize < ACHAT_HEADER_SIZE) {
    printf ("useless decrypted chat message of size %d, header size %d\n",
            bsize, ACHAT_HEADER_SIZE);
    return;
  }
  if (rsize > size) {
    printf ("decoded message %d larger than received %d\n", rsize, size);
    rsize = size;
  }
  struct achat_header * achp = (struct achat_header *) text_buffer;
  char * data = text_buffer + ACHAT_HEADER_SIZE;
  int dsize = rsize - ACHAT_HEADER_SIZE;

  /* send the ack */
  allnet_send_ack (achp->hashes_to_packet_id,

  free (text_buffer);



  int offset = PACKET_ID_SIZE + PACKET_ID_SIZE;
  if (bsize < offset + COUNTER_TIME_SIZE) {
    printf ("minimum message size %d, received %d\n",
            offset + COUNTER_TIME_SIZE, bsize);
    print_buffer (buffer, bsize, 1);
#if 0
  /* should ignore it, but for now, ack it, may be an old format */
  int asize;
  char * ack = allnet_ack (buffer, bsize, &asize);
  route_num_addr = 1;
  route_used = 0;
  struct allnet_addr dest;
  dest.discrim = ALLNET_BCAST_ADDR_ID;
  dest.bc_addr.hops = 6;
  allnet_send_clear (ack, asize, &dest, route_num_addr, route_used);
  free (ack);
  /* end ack */
#endif /* 0 */
    return;
  }

/*
  printf ("handle_user_message (%p, %d, ... %d)\n",
          buffer, bsize, key_id);
*/
  int asize;
  char * ack = allnet_ack (buffer, bsize, &asize);
/* this simply puts my address back in, which is useless -- fix this
 * once allnetd actually records the route.
 */
  reverse_route (route, route_num_addr, &route_used, route);
/*
  route_num_addr = 1;
  route_used = 0;
  struct allnet_addr dest;
  dest.discrim = ALLNET_BCAST_ADDR_ID;
  dest.bc_addr.hops = 6;
  allnet_send_clear (ack, asize, &dest, route_num_addr, route_used);
*/
  allnet_send_clear (ack, asize, route, route_num_addr, route_used);
  free (ack);

  message_counter sent_c;
  message_time sent_time;
/*
  printf ("chat saving received message (+ %d):\n", offset);
  print_buffer (buffer, bsize, 1);
*/
  save_received_message (key_id, buffer + offset, bsize - offset,
                         &sent_c, &sent_time);
  bsize -= COUNTER_TIME_SIZE;  /* removing counter and time */

  char string [MAGIC_SIZE];
  int size = MAGIC_SIZE - 1;
  if (bsize - offset < size)
    size = bsize - offset;
  memcpy (string, buffer + offset, size);
  string [size] = '\0';

/*
  printf ("received message sent at time %ld, offset %d, now is %ld\n",
          sender_gtime, sender_time_offset, time (NULL));
*/
  char to_print [MAGIC_SIZE * 3];
  msg_to_string (key_id, sent_c, sent_time, string,
                 to_print, sizeof (to_print));
  char time_string [MAGIC_SIZE];
  time_to_string (time_string, MAGIC_SIZE, sent_time);

  char name [MAGIC_SIZE] = "unknown";
  char user_path [MAGIC_SIZE] = "";
  name_for_key (key_id, name, MAGIC_SIZE - 1, user_path, MAGIC_SIZE - 1);

  snprintf (to_print, sizeof (to_print),
            "(%lld) from %s at %s: %s", sent_c, name, time_string, string);

  if (! initializing)
    printf ("\n  %s\n%s", to_print, prompt);
  else
    save_message (to_print, sent_time, sent_c, key_id);
}

static void handle_key_message (char * buffer, int bsize,
                                struct allnet_header * hp)
{
  if ((received_key_count == -1) ||  /* not expecting anything */
      (received_key_count > 100)) {  /* too many tries */
    if (received_key_count != -1)
      printf ("received %d keys, ignoring this one\n", received_key_count);
    received_key_count = -1;
    return;
  }
  int verification = verify_key_exchange (buffer, bsize, expecting_secret,
                                          expecting_peer);
  if (verification >= 0) {
    printf ("\nreceived and saved key for %s\n", expecting_peer);
    received_key_count = -1;
    /* send the response with 2 more hops than we got it with, to be safe */
    send_key_hash (expecting_secret, exchanging_key_id, hp->hops_so_far + 2);
  } else {
    if (verification == -2) {
      /* printf ("received my own key message, ignoring\n"); */
    } else {
      printf ("message of length %d, verification %d\n", bsize, verification);
      received_key_count++;
    }
  }
}

static void handle_ack_message (char * buffer, int bsize,
                                struct allnet_header_data * dhp)
{
  if ((waiting_for_ack) &&
      (is_valid_ack (buffer, bsize, dhp->packet_id, PACKET_ID_SIZE))) {
    waiting_for_ack = 0;
    pthread_cond_broadcast (&ack_cond);
  } /* else ignore the ack */
}

void receive_fun (void * arg, char * buffer, int bsize)
{
  /* printf ("received packet of size %d\n", bsize); */
  /* caller should make sure packet has a valid header */
  struct allnet_header * hp = (struct allnet_header *) buffer;
  struct allnet_data_header * dhp = (struct allnet_data_header *) buffer;
  switch (hp->packet_type) {
  case define ALLNET_TYPE_DATA:
    break;
  case ALLNET_TYPE_DATA_ACK:
    handle_ack_message (buffer, bsize, dhp);
    break;
  case ALLNET_TYPE_DATA_REQ:
    break;
  case ALLNET_TYPE_KEY_XCHG:
    handle_key_message (buffer, bsize, hp);
    break;
  case ALLNET_TYPE_CLEAR:
    break;
  case ALLNET_TYPE_MGMT:
    break;
  default:
    printf ("achat discarding packet with unknown type %d/0x%x\n",
            hp->packet_type, hp->packet_type);
    return;
  }
}

static void send_message (char * user_input, char * user, int current_key,
                          int counter)
{
  if ((current_key == -1) || (counter == -1)) {
    printf ("no user selected, unable to send message\n");
  }

  char actual_message [MAGIC_SIZE * 3];
  int offset = PACKET_ID_SIZE;
  int send_len = 0;
  send_len = snprintf (actual_message + offset,
                       sizeof (actual_message) - offset - COUNTER_TIME_SIZE,
                       "%s", user_input);
  if (send_len >= 0) {
    save_outgoing_message (current_key, actual_message + offset, send_len);
    send_len += offset + COUNTER_TIME_SIZE;
/*
    printf ("outgoing message is (%d bytes):\n", send_len);
    print_buffer (actual_message + offset, send_len - offset, 1);
*/
  }
  random_fill (sent_message_owner_id, sizeof (sent_message_owner_id), 1, 0);
  waiting_for_ack = 1;
  /* only really interested in the signal, not the mutex */
  pthread_mutex_t useless_mutex = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock (&useless_mutex);
  struct timespec finish_time;
  clock_gettime (CLOCK_REALTIME, &finish_time);
  finish_time.tv_sec += 3;   /* wait up to 3s for ack */

/*
  printf ("sending header: ");
  print_buffer (actual_message, 12, 1);
*/

/*
  printf ("encrypting buffer: ");
  print_buffer (actual_message, send_len, 1);
*/
  struct allnet_addr dest;
  dest.discrim = ALLNET_KEY_ADDR_ID;
  dest.pk_addr.hops = 6;
  dest.pk_addr.num_bits = 8;  /* may change later */
  int osize = sizeof (sent_message_owner_id);
  int bsize;
  char * buf = allnet_encrypt (actual_message, send_len, current_key,
                               sent_message_owner_id, osize, &bsize,
                               b (dest.pk_addr.num_bits), dest.pk_addr.pubkey);
  allnet_send (buf, bsize, &dest, 1, 0, NULL, 0);

  if (pthread_cond_timedwait (&ack_cond, &useless_mutex, &finish_time) == 0)
    pthread_mutex_unlock (&useless_mutex);
  if (waiting_for_ack == 0) {
    printf ("  (acknowledged)\n");
  } else { /* not acked, re-send it to whoever will store it */
    allnet_send (buf, bsize, &dest, 1, 0, sent_message_owner_id, osize);
  }
  waiting_for_ack = 0;
  free (buf);
}

static void send_key ()
{
  char p [MAGIC_SIZE];
  snprintf (p, sizeof (p), "enter the secret string of your new contact: ");
  char * shared_secret = my_readline (p, &received_key_count, "key received");
  if (shared_secret == NULL) {
    /* printf ("no string entered\n"); */
    return;
  }
  if ((*shared_secret == '\0') || (*shared_secret == '0')) {
    /* printf ("no string entered\n"); */
    free (shared_secret);
    return;
  }
  strip_non_alpha (shared_secret);
  make_upper (shared_secret);
  strncpy_t (expecting_secret, shared_secret, MAGIC_SIZE);

  send_key_hash (shared_secret, 1);
  send_key_hash (shared_secret, 6);
}

static void receive_key ()
{
  int have_name = 0;
  char * peer_name =
    my_readline ("enter the name to use for your new contact: ", NULL, NULL);
  if ((peer_name == NULL) || (*peer_name == '\0')) {
    time_t now = time (NULL);
    snprintf (expecting_peer, MAGIC_SIZE, "key received on %s", ctime (&now));
    if (expecting_peer [strlen (expecting_peer) - 1] == '\n')
      expecting_peer [strlen (expecting_peer) - 1] = '\0';
  } else {
    strncpy_t (expecting_peer, peer_name, MAGIC_SIZE);
    have_name = 1;
  }
  if (peer_name != NULL)
    free (peer_name);
  received_key_count = 0;  /* get ready to receive */

  char * use_name = "your new contact";
  if (have_name)
    use_name = expecting_peer;

  new_keypair (use_name);

#define SECRET_LENGTH	(12 + 2)
  char random_secret [MAGIC_SIZE];
  random_fill (random_secret, SECRET_LENGTH, 1, 1);
  random_secret [SECRET_LENGTH / 3] = ' ';
  random_secret [2 * SECRET_LENGTH / 3] = ' ';
  random_secret [SECRET_LENGTH] = '\0';
#undef SECRET_LENGTH
  char lower_secret [MAGIC_SIZE];
  strcpy (lower_secret, random_secret);
  strcpy (expecting_secret, random_secret);
  strip_non_alpha (expecting_secret);
  make_upper (expecting_secret);
  make_lower (lower_secret);

  printf ("either you or %s must enter a secret string\n", use_name);
  printf ("if you enter it, enter the one shown on the screen of %s\n",
          use_name);
  printf ("if %s enters it, tell them to enter '%s' (or %s)\n",
          use_name, expecting_secret, lower_secret);
  send_key ();
}

static void my_exit ()
{
  allnet_stop ();
  printf ("goodbye\n");
  exit (0);
}

#define MENU_EXIT		0
#define MENU_NO_ACTION		1
#define MENU_NEW_USER		2
#define MENU_MESSAGE		7
#define MENU_EXCHANGE_KEY	9
static int main_menu (char * user, /* pointer to MAGIC_SIZE bytes */
                      char ** message) /* set if returning MENU_MESSAGE
                                          or MENU_NEW_USER */
{
  if (*user == '\0')
    snprintf (prompt, sizeof (prompt),
              "enter 0 to quit, 9 to exchange keys, or user name: ");
  else
    snprintf (prompt, sizeof (prompt),
              "0 quit, 9 exchange keys, or message for %s: ", user);
/*
    snprintf (prompt, sizeof (prompt),
              "0 quit, 1 new user, 9 exchange keys, or message for %s: ", user);
*/
  char * input = my_readline (prompt, NULL, NULL);
  if (input == NULL)
    return MENU_EXIT;
  if (strcmp (input, "0") == 0) {
    free (input);
    return MENU_EXIT;
  }
  if (strcmp (input, "9") == 0) {
    free (input);
    return MENU_EXCHANGE_KEY;
  }
  if (*user == '\0') {
    if (input == NULL)
      return MENU_NO_ACTION;
    if (*input == '\0') {
      free (input);
      return MENU_NO_ACTION;
    }
    if (*input == '0') {
      free (input);
      return MENU_EXIT;
    }
    /* name entered */
    *message = input;
    return MENU_NEW_USER;
  } /* else it is a message for user */
  *message = input;
  return MENU_MESSAGE;
}

static void set_user (char * new_name, int * key, message_counter * counter,
                      char * nick, int nsize, char * user_name, int usize,
                      char * user_path, int upsize)
{
  if (key_for_name (new_name, key, 1) <= 0) {
    printf ("user %s has no key\n", new_name);
    *key = -1;
    *counter = ((message_counter) (-1));
    *nick = '\0';
  } else {
    strncpy_t (nick, new_name, nsize);
    *counter = present_counter (*key);
    if (name_for_key (*key, user_name, usize, user_path, upsize) > 0) {
      printf ("talking to user %s, key ID %d\n", user_name, *key);
    } else {  /* got key from key_for_name, it should be valid */
      printf ("error: name_for_key failed for key ID %d, nickname %s\n",
              *key, nick);
      exit (1);  /* figure out what happened */
    }
  }
#define PRINT_PAST_MSGS		20
  struct message_buffer print [PRINT_PAST_MSGS];
  int num = last_n_messages (*key, print, PRINT_PAST_MSGS, -1);
  int i = 0;
  for (i = 0; i < num; i++) {
    char string [MAGIC_SIZE * 3];
    msg_to_string (*key, print [i].counter, print [i].time, print [i].message,
                   string, sizeof (string));
    if (print [i].sent)
      printf ("  -> %s\n", string);
    else
      printf ("<- %s\n", string);
    free (print [i].message);
  }
#undef PRINT_PAST_MSGS
}

static void usage (char * programname)
{
  printf ("usage: %s [-c configuration-directory] name\n", programname);
  printf ("  for example, to talk to Bill: %s Bill\n", programname);
  printf ("  to use a configuration directory named 'test': %s -c test Bill\n",
          programname);
  printf ("  if the name is missing, it can be provided later\n");
  exit (1);
}

main (int argc, char ** argv)
{
  /* should be done first, before any use of stdout */
  if (setvbuf (stdout, NULL, _IONBF, 0) != 0) {
    perror ("setvbuf");
    printf ("unable to set stdout to unbuffered mode\n");
  }

  char * home = getenv ("HOME");
  char path [MAGIC_SIZE];
  snprintf (path, MAGIC_SIZE, "%s/%s", home, ALLNET_API_DIRECTORY);
  if ((argc > 2) && (strcmp (argv [1], "-c") == 0)) {
    strncpy_t (path, argv [2], MAGIC_SIZE);
    argc -= 2;
    argv += 2;
  }

  /* int port = allnet_start (receive_fun, NULL); */

  int num_keys = allnet_start (receive_fun, NULL, path, "chat");
  if (num_keys < 0) {
    printf ("unable to open key files, aborting\n");
    exit (1);
  }

  struct allnet_addr route [2];
  route [0].discrim = ALLNET_KEY_ADDR_ID;
  route [0].pk_addr.hops = 1;
  route [0].pk_addr.num_bits = 16;
  route [1].discrim = ALLNET_BCAST_ADDR_ID;
  route [1].bc_addr.hops = 6;
  /* request any outstanding messages */
  /* should put my key here, send with a pub key source address TO
   * a broadcast address. */
  /* note: messages are returned in priority order, meaning most recent
   * first.  Really should display oldest first */
  int old_saved_messages;
  do {
    old_saved_messages = saved_messages;
    int i;
    char message [CLEAR_ID_SIZE * 3];
    memcpy (message, REQUEST_BYTES, CLEAR_ID_SIZE);
    /* add some randomness to the message, otherwise it might look
     * like a message sent earlier */
    random_fill (message + CLEAR_ID_SIZE, CLEAR_ID_SIZE * 2, 1, 0);
/* does not scale, fix later */
    for (i = 0; i < num_keys; i++) {
      key_bytes (i, route [0].pk_addr.pubkey, 2);
      allnet_send_clear (message, sizeof (message), route, 2, 1);
    }
    sleep (1);
    if (old_saved_messages != saved_messages)
      printf ("got %d saved messages\n", saved_messages);
  } while (old_saved_messages != saved_messages);
  initializing = 0;
  int i, ik;
  for (ik = 0; ik < num_keys; ik++) {
    for (i = saved_messages - 1; i >= 0; i--) {
      if (messages [i].key_id == ik) {
        printf ("%s\n", messages [i].string);
      }
    }
  }

  char nick [MAGIC_SIZE] = "";
  char user_name [MAGIC_SIZE] = "";
  char user_path [MAGIC_SIZE] = "";
  int current_key = -1;
  message_counter counter = (message_counter) (-1);
  if (argc > 1)
    set_user (argv [1], &current_key, &counter, nick, sizeof (nick),
              user_name, sizeof (user_name), user_path, sizeof (user_path));
  printf ("send counter set to %lld\n", counter);
  while (1) {
    char * message;
    int request = main_menu (nick, &message);
    switch (request) {
      case MENU_EXIT:
        my_exit ();
        break;             /* in case my_exit ever returns */
      case MENU_NO_ACTION:
        break;
      case MENU_NEW_USER:  /* user should have been set, set key */
        set_user (message, &current_key, &counter, nick, sizeof (nick),
                  user_name, sizeof (user_name), user_path, sizeof (user_path));
        printf ("send counter set to %lld\n", counter);
        break;
      case MENU_MESSAGE:
        send_message (message, nick, current_key, counter);
        counter++;
        break;
      case MENU_EXCHANGE_KEY:
        receive_key ();
        break;
      default:
        printf ("unimplemented menu item %d\n", request);
        break;
    }
  }
}
