/* achat.h: definition of achat packet formats */

#ifndef ACHAT_H
#define ACHAT_H

#include "packet.h"

/* every chat message begins with an achat_header followed by the data. */
/* the whole message is encrypted with the recipient's public key */

/* achat_header begins with PACKET_ID_SIZE bytes of nonce, which when
 * hashed, should give the packet ID */
/* next comes an 8-byte (64-bit) counter, intended to be unique, for
 * any pair of correspondents, over all time */
/* next comes the time, also 64 bits, in seconds since 00:00 Jan 1, 2000 GMT */
/* all this is followed by the data, the size of which is variable */

/* time allows editing of messages.  A recipient of a message with the
 * same counter and later time should keep both, but normally only display
 * the most recent, preferably with an indication that older versions are
 * available.  If the newest message has no data, that is an indication
 * that the sender wishes to delete the message.
 */

/* the counter allows the receiver of a later counter value to request
 * messages with missing counter values.
 */

/* the control word, if nonzero, specifies that this is not a text
 * message but a specific request.  Intended is that this should
 * support transfers of large data (e.g. multimedia), as well as
 * perhaps allow establishment of sockets, if necessary through
 * intermediaries with routable IP addresses.
 * Also for exchanging profile information.
 * But this is all TBD.
 */

#define CHAT_COUNTER_BYTES	8
#define CHAT_TIME_BYTES		8
#define CHAT_REQUEST_BYTES	4

struct achat_header {
  unsigned char hashes_to_packet_id [PACKET_ID_SIZE];
  unsigned char counter [CHAT_COUNTER_BYTES];
  unsigned char time [CHAT_TIME_BYTES];
  unsigned char request [CHAT_REQUEST_BYTES];
};

#define ACHAT_HEADER_SIZE	(sizeof (struct achat_header))

#endif /* ACHAT_H */
