/* alocal.c: forward allnet messages to and from local clients */
/* two threads:
 * - one listens for connections on localhost port 0xa11e (ALLnEt)
 * - one listens for messages from ad or from clients, and forwards them
 *   to clients and ad
 * alocal takes two arguments, the fd of a pipe from AD and of a pipe to AD
 */
/* there are actually two listen threads, one for IPv4 and one for IPv6 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "packet.h"
#include "pipemsg.h"
#include "listen.h"

void * main_loop (int rpipe, int wpipe, struct listen_info * info)
{
  while (1) {
    int fd;
    int priority;
    char * message;
    int result = receive_pipe_message_any (1000, &message, -1, &fd, &priority);
    if (result < 0) {
      if (fd == rpipe) {
        printf ("alocal ad pipe %d closed\n", rpipe);
        break;
      }
      printf ("alocal: error on file descriptor %d, closing\n", fd);
      listen_remove_fd (info, fd);
      close (fd);       /* remove from kernel */
    } else if (result > 0) {
      if (fd == rpipe) {    /* message from ad, send to all clients */
        int i;    /* start with i = 1, no sending to ad read pipe */
        pthread_mutex_lock (&(info->mutex));
        for (i = 1; i < info->num_fds; i++) {
          if (! send_pipe_message (info->fds [i], message, result, priority)) {
            printf ("alocal error sending to info pipe %d at %d\n",
                    info->fds [i], i);
            /* listen_remove_fd (info, info->fds [i]); */
          } else {
            printf ("sent to info pipe %d at %d %d bytes, priority %08x\n",
                    info->fds [i], i, result, priority);
          }
        }
        pthread_mutex_unlock (&(info->mutex));
      } else {              /* message from a client, send to ad */
        listen_record_usage (info, fd);  /* make it most recently used */
        if (! send_pipe_message (wpipe, message, result, priority)) {
          printf ("alocal error sending to ad pipe %d\n", wpipe);
          break;
        }
        printf ("alocal sent %d bytes to ad pipe %d\n", result, wpipe);
      }
      free (message);
    }   /* else result is zero, timed out, try again */
  }
}

int main (int argc, char ** argv)
{
  if (argc != 3) {
    printf ("arguments must be a read and a write pipe\n");
    return -1;
  }
/*
  printf ("in alocal, args are ");
  printf ("'%s %s %s'\n", argv [0], argv [1], argv [2]);
*/
  int rpipe = atoi (argv [1]);
  int wpipe = atoi (argv [2]);
  /* printf ("read pipe is fd %d, write pipe is fd %d\n", rpipe, wpipe); */
  struct listen_info info;
  listen_init_info (&info, 256, "alocal", ALLNET_LOCAL_PORT, 1, 1);
  listen_add_fd (&info, rpipe);

  main_loop (rpipe, wpipe, &info);
  pthread_cancel (info.thread4);
  pthread_cancel (info.thread6);
  printf ("end of alocal main thread\n");
}
