/* listen.c: thread to listen on a port and maintain connected fds */
/*   there is a finite maximum number of fds -- once more are connected, */
/*   old ones are closed */

#ifndef LISTEN_H
#define LISTEN_H

/* this structure is declared in the caller and passed to every function.
 * Some of the fields should not be accessed by the caller */
struct listen_info {
  /* accessed by the caller as well as the code in listen.c */
  int num_fds;           /* counter for number of file descriptors in fds */
  int * fds;             /* array of ints for file descriptors */
  pthread_mutex_t mutex; /* mutex for accessing fds */
  pthread_t thread4;     /* listen thread for IPv4 */
  pthread_t thread6;     /* listen thread for IPv6 */
  /* the rest of these fields are for listen.c internal use only */
  int max_num_fds;       /* max number of elements with space in fds */
  char * program_name;   /* e.g. alocal, aip */
  int listen_fd4;        /* fd for listening for new connections on ipv4 */
  int listen_fd6;        /* fd for listening for new connections on ipv6 */
  int port;              /* TCP port number */
  int add_remove_pipe;   /* call add_pipe and remove_pipe */
  /* for testing, make counter a char.  Normally, unsigned int */
  unsigned char counter;  /* cycle counter for least recently used */
  unsigned int * used;   /* array of most recent access times */
};

/* exits in case of errors, otherwise initializes info and starts the
 * listen thread */
/* ip version should be 4 or 6 */
/* add_remove_pipe should be 1 if add_pipe and remove_pipe should be
 * called when adding or removing pipes */
extern void listen_init_info (struct listen_info * info, int max_fds,
                              char * name, int port, int local_only,
                              int add_remove_pipe);

/* call to record that this fd was active at this time */
extern int listen_record_usage (struct listen_info * info, int fd);

/* call to add an fd to the data structure */
/* may close the least recently active fd */
extern void listen_add_fd (struct listen_info * info, int fd);

/* call to remove an fd from the data structure */
extern void listen_remove_fd (struct listen_info * info, int fd);

#endif /* LISTEN_H */
