/* fraction.c: compute with fractions using integers */

#include <stdio.h>

#include "priority.h"

void print_fraction (int value, char * str)
{
  double v = ((double) value) / ((double) MAX_PRIORITY);
  if (str != NULL)
    printf ("%s: %f\n", str, v);
  else
    printf ("%f", v);
}

int power_half_fraction (int power)
{
  if (power <= 0)
    return MAX_PRIORITY;
  if (power >= 30)
    return 0;
  return MAX_PRIORITY >> power;
}

/* multiplying fractions, so compute with 64 bits, then shift down 30 bits */
int multiply (int p1, int p2)
{
  long long int product = ((long long int) p1) * ((long long int) p2);
  int result = product >> 30;
  return result;
}

int compute_priority (int is_local, int size, int dest_bits,
                      int hops_already, int hops_left,
                      int social_distance, int rate_fraction)
{
  int debug = 0;
  if (debug)
    printf ("compute_priority (%slocal, %d, %d, %d, %d, %d, %d)\n",
            (is_local ? "" : "non-"), size, dest_bits, hops_already,
            hops_left, social_distance, rate_fraction);
  if (is_local)
    return SEVEN_EIGHTS;  /* sender may have specified another priority */
  if (social_distance <= 1)
    return THREE_QUARTERS;
  /* compute Ps = 2^(-social_distance + 1).
   * So for d == 2, Ps = 0.5, for d == 3, Ps = 0.25, etc */
  int social_priority = power_half_fraction (social_distance - 1);
  if (debug) print_fraction (social_priority, "social");

  /* For Pg and Ph 1/g and 1/h heavily prioritizes local traffic */
  int hops_carried_priority = MAX_PRIORITY / hops_already;
  if (debug) print_fraction (hops_carried_priority, "hops_carried");
  int hops_left_priority = MAX_PRIORITY / hops_left;
  if (debug) print_fraction (hops_left_priority, "hops_left");

  /* compute Pb as 1 - 2^-(dest_bits+1).  So for dest_bits == 0, Pb = 1/2,
     for dest_bits = 1, Pb = 3/4, for dest_bits = 2, Pb = 7/8, etc. */
  int bits_priority = MAX_PRIORITY - power_half_fraction (dest_bits + 1);
  if (debug) print_fraction (bits_priority, "bits");

  /* Pl = 1 - r' / r, where r' is our sending rate, and r is the maximu
   * sending rate.  rate_fraction is already r' / r */
  int rate_priority = MAX_PRIORITY - rate_fraction;
  if (rate_priority == 0) rate_priority = 1;
  if (debug) print_fraction (rate_priority, "rate");

  /* combine these as 1 - (1 - Ps) * (1 - Pb * Pg * Ph * Pl) */
  int result = multiply (social_priority,
                         multiply (multiply (bits_priority, rate_priority),
                                   multiply (hops_carried_priority,
                                             hops_left_priority)));
  if (debug) print_fraction (result, "result");
  
  return result;
}

