/* sha.c: compute sha512 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <openssl/evp.h>    /* for hash computation */

#include "sha.h"

/* the result array must have size SHA512_SIZE */
/* #define SHA512_SIZE	64 */
void sha512 (char * data, int dsize, char * result)
{
  static EVP_MD_CTX sha512_ctx;
  static const EVP_MD * sha512 = NULL;

  if (sha512 == NULL) {
    OpenSSL_add_all_digests ();  /* supposedly required */
    sha512 = EVP_get_digestbyname ("sha512");
    if (EVP_MD_size (sha512) != SHA512_SIZE) {
      printf ("error: expected hash result size is %d, %d expected\n",
              EVP_MD_size (sha512), SHA512_SIZE);
      exit (1);
    }
  }
  EVP_DigestInit (&sha512_ctx, sha512);
  EVP_DigestUpdate (&sha512_ctx, data, dsize);
  unsigned int result_size;
  EVP_DigestFinal (&sha512_ctx, result, &result_size);
  if (result_size != SHA512_SIZE)
    printf ("error: hash result size is %ud, %d expected\n",
            result_size, SHA512_SIZE);
}

