
#include <stdio.h>
#include <stdlib.h>

#include "social.h"
#include "table.h"
#include "config.h"

struct social_one_tier {
  int bytes_per_entry;
  /* the IDs of the contacts, in a table */
  struct table connections;
};

static void print_social_tier (int tier, struct social_one_tier * soc)
{
  printf ("tier %d: %d contacts, %d slots, total size %d\n", tier,
          soc->connections.num_entries, soc->connections.num_slots,
          soc->connections.storage_size);
/*
  printf ("  %d entries with %d user bytes give %d bytes, total %d bytes\n",
          soc->connections.num_entries, soc->connections.bytes_per_entry,
          soc->bytes_per_entry, soc->connections.storage_size);
  int i;
  for (i = 0; i < soc->connections.num_slots; i++)
    print_table_entry (i, soc->connections.table [i],
                       soc->connections.num_data_bytes);
 */
}

/* keep track of up to friends (3) of friends (2) of friends (1)
 * indices are 0 (for friends), 1 (friends of friends), 2 (friends of friends
 * of friends)
 define MAX_SOCIAL_TIER		3
*/

struct social_info {
  struct social_one_tier info [MAX_SOCIAL_TIER];
  int max_bytes;    /* should not use more than max_bytes of storage */
};

struct social_info * init_social (int max_bytes)
{
  struct social_info * result = malloc (sizeof (struct social_info));
  if (result == NULL) {
    perror ("init_social(0) malloc");
    printf ("unable to allocate %zd bytes for social info\n",
            sizeof (struct social_info));
    exit (1);
  }
  result->max_bytes = max_bytes;
  int bytes = 512;
  int i;
  for (i = 0; i < MAX_SOCIAL_TIER; i++) {
    init_table (&result->info [i].connections);
    result->info [i].bytes_per_entry = bytes;
    if (bytes >= 512) {        /* decrease the number of bytes per entry */
      bytes = 16;
    } else if (bytes >= 16) {
      bytes = 8;
    } else {
      bytes--;
    }
  }
  return result;
}

/* return the number of bytes in the updated social tier, and in
 * any case never more than free_entries */
static int update_social_tier (int tier, struct social_one_tier * st,
                               int free_bytes)
{
  if (tier > 99) {
    printf ("error, using social tier %d, maximum tier is 99\n", tier);
    if (st->connections.storage != NULL)
      free (st->connections.storage);
    init_table (&(st->connections));
    return 0;
  }
  char file_name [] = "social12";  /* two digits allow up to 99 tiers */
  snprintf (file_name, sizeof (file_name), "social%d", tier);
  static int printed = 0;
  int fd = open_read_config ("ad", file_name, ! printed);
  if (fd < 0) {
    char * path;
    config_file_name ("ad", file_name, &path);
    if (! printed) {
      if (fd == -1)   /* no such file */
        printf ("no social info file %s for ad, continuing\n", path);
      else
        printf ("error reading social info file %s for ad\n", path);
    }
    printed = 1;
    free (path);
    return free_bytes;
  }
  int bytes = table_from_file (&(st->connections), fd,
                               st->bytes_per_entry, free_bytes);
  close (fd);
  if (bytes != 0)
    return bytes;
  else   /* did not read from file, using older one */
    return st->connections.storage_size;
}

time_t update_social (struct social_info * soc, int update_seconds)
{
  int next_seconds = time (NULL) + update_seconds;
  int free_bytes = soc->max_bytes;
  int i;
  for (i = 0; i < MAX_SOCIAL_TIER; i++) {
    free_bytes -= update_social_tier (i, soc->info + i, free_bytes);
    print_social_tier (i, soc->info + i);
  }
  return next_seconds;
}

int social_connection (struct social_info * soc, char * dest, int bits)
{
  int result = COMPLETE_STRANGER;
  if (bits <= 8)   /* less than 8 bits, no use even checking */
    return result;
  int i;
  for (i = 0; i < MAX_SOCIAL_TIER; i++) {
    if (table_find (dest, bits, &(soc->info [i].connections))) {
      result = i + 1;
      break;
    }
  }
  /* for every 4 additional bits beyond 8, allow it to be one degree closer */
  int best_allowed = COMPLETE_STRANGER - (bits - 5) / 4;
  if (best_allowed < 1) best_allowed = 1;   /* one means a friend */
  /* printf ("result %d, best allowed %d for %d bits\n", result, best_allowed, bits); */
  if (result < best_allowed) result = best_allowed;
  return result;
}
