/* table.h: store arbitrary bitstrings */

#ifndef ALLNET_TABLE_H
#define ALLNET_TABLE_H

struct table {
  int num_entries;
  int num_slots;	 /* the first power of two that includes num_entries */
  int num_data_bytes;    /* bytes of user data in each entry */
  int bytes_per_entry;   /* bytes allocated in each entry */
  struct table_entry ** table;
  int storage_size;
  char * storage;
  char * freeptr;
};

/* will not exceed free_bytes, ignoring entries that would require
 * too much room */
/* ignores entries that have more or fewer than bytes_per_entry */
/* returns the number of bytes in the table. */
/* in case of failure returns 0 and the table is unchanged */
extern int table_from_file (struct table * table,
                            int fd, int bytes_per_entry, int free_bytes);

/* returns 1 if found, 0 otherwise */
extern int table_find (char * bitstring, int bits,
                      struct table * table);

extern void init_table (struct table * table);

/* for debugging */
extern void print_table_entry (int index, struct table_entry * entry,
                               int num_data_bytes);

#endif /* ALLNET_TABLE_H */
