/* tac.c: test acache.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "../packet.h"
#include "../pipemsg.h"
#include "../priority.h"
#include "../util.h"
#include "../sha.h"

/* copied from tsocial.c */
static void init_array (char * data, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    data [i] = (first + i) & 0xff;
  print_buffer (data, count, "init", 18, 1);
}

static void init_pipes (int * pipes, int num_pipes)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (pipe (pipes + i * 2) < 0) {
      perror ("pipe");
      printf ("error creating pipe set %d\n", i);
      exit (1);
    }
  }
}

static void init_packet (char * p, int psize, int type,
                         int hops, int max_hops, int nbits, int first)
{
  p [0] = type;
  p [1] = hops;
  p [2] = max_hops;
  p [3] = nbits;
  init_array (p + 4, first, psize - 4);
}

static void receive_all_messages (int * pipes, int count)
{
  int loops = 0;
  while (1) {
    char * message;
    int pipe, pri;
    int r = receive_pipe_message_any (1000, &message, -1, &pipe, &pri);
    if (r <= 0) {
      if (loops > 0) printf ("no more messages\n");
      else           printf ("no messages\n");
      break;
    }
    loops++;
    printf ("received %d bytes on pipe %d", r, pipe);
    printf (", priority ");
    print_fraction (pri, NULL);
    printf ("\n");
    print_buffer (message, r, NULL, 20, 1);
    free (message);
  }
}

static int new_local_socket (int ipv6)
{
  int fd = socket (ipv6 ? AF_INET6 : AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
    perror ("socket");
    exit (1);
  }
  struct sockaddr_storage address;
  struct sockaddr     * ap  = (struct sockaddr     *) &address;
  struct sockaddr_in  * ap4 = (struct sockaddr_in  *) ap;
  struct sockaddr_in6 * ap6 = (struct sockaddr_in6 *) ap;
  socklen_t addr_size = sizeof (address);

  if (ipv6) {
    ap6->sin6_family = PF_INET6;
    memcpy (&(ap6->sin6_addr), &(in6addr_loopback), sizeof (ap6->sin6_addr));
    ap6->sin6_port = ALLNET_PORT;
  } else {
    ap4->sin_family = AF_INET;
    ap4->sin_addr.s_addr = htonl (INADDR_LOOPBACK);
    ap4->sin_port = ALLNET_PORT;
  }
  
  if (connect (fd, ap, addr_size) < 0) {
    perror ("connect");
    exit (1);
  }
  return fd;
}

static void parent (int * pipes)
{
  close (pipes [0]);
  close (pipes [3]);
  int write_pipe = pipes [1];
  int read_pipe = pipes [2];

  printf ("in parent\n");
  add_pipes (&read_pipe, 1);
  sleep (1);

  char packet [200];
  int now = time (NULL);
  /* send a data packet from ad, nothing should really happen... */
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 1, 2, 16, now); 
  char packet_id [PACKET_ID_SIZE];
  char packet_id_sha [SHA512_SIZE];
  init_array (packet_id, now +156, PACKET_ID_SIZE);
  sha512 (packet_id, PACKET_ID_SIZE, packet_id_sha);
  memcpy (packet + ALLNET_BASIC_HEADER_SIZE, packet_id_sha, PACKET_ID_SIZE);
  printf ("sending data packet from ad, 1 -> 2 hops, 1/4 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet), ONE_QUARTER);
  receive_all_messages (&read_pipe, 1);
  sleep (2);

  /* now send from ad a request for that same packet */
  char req [140];
  init_packet (req, sizeof (req), ALLNET_TYPE_DATA_REQ, 0, 1, 16, now + 73); 
  req [ALLNET_BASIC_HEADER_SIZE] = 8;  /* 8 bits of match */
  init_array (req + ALLNET_BASIC_HEADER_SIZE + 1, now,
              sizeof (req) - (ALLNET_BASIC_HEADER_SIZE + 1));
  printf ("sending request packet from ad, 0 -> 1 hops, 1/2 priority\n");
  send_pipe_message (write_pipe, req, sizeof (req), ONE_HALF);
  receive_all_messages (&read_pipe, 1);
  sleep (2);

  /* now send from ad a request for a different packet */
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA_REQ, 0, 1, 16,
               now + 42); 
  req [ALLNET_BASIC_HEADER_SIZE] = 16;  /* 16 bits of match */
  init_array (packet + ALLNET_BASIC_HEADER_SIZE + 1, now + 1,
              sizeof (packet) - (ALLNET_BASIC_HEADER_SIZE + 1));
  printf ("sending request for a packet we should not have\n");
  send_pipe_message (write_pipe, packet, sizeof (packet), ONE_HALF);
  receive_all_messages (&read_pipe, 1);
  sleep (2);

  /* now send a request for the packet that should still be there */
  printf ("sending good request packet from ad\n");
  send_pipe_message (write_pipe, req, sizeof (req), ONE_HALF);
  receive_all_messages (&read_pipe, 1);
  sleep (2);

  /* now send an ack for the packet that is saved */
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA_ACK, 1, 7, 16,
               now + 201); 
  /* identify which packet we are acknowledging */
  memcpy (packet + ALLNET_BASIC_HEADER_SIZE, packet_id, PACKET_ID_SIZE);
  printf ("sending ack packet from ad, 1 -> 7 hops, 1/2 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet), ONE_HALF);
  receive_all_messages (&read_pipe, 1);
  sleep (2);

  /* now send the request again, the packet should be gone */
  printf ("sending request packet from ad for packet that was acked\n");
  send_pipe_message (write_pipe, req, sizeof (req), ONE_HALF);
  receive_all_messages (&read_pipe, 1);
  sleep (2);

  printf ("tac main thread exiting\n");
}

static char * itoa (int n)
{
  char * result = malloc (12);  /* plenty of bytes, easier than being exact */
  snprintf (result, 10, "%d", n);
  return result;
}

static void prt_str (int fd, const char * str)
{
  while (*str != '\0')
    write (fd, str++, 1);
}

static void print_prefix (const char * prefix, int replace_fd)
{
  /* for redirecting the output to a terminal, do "who am i" in that terminal,
     and put the terminal name here */
/*
  int tty_fd = open ("/dev/pts/3", O_WRONLY);
  dup2 (tty_fd, 1);
  return;
*/
  int stdoutpipe [2];
  pipe (stdoutpipe);
  if (fork () == 0) {   /* child, print prefix */
    close (stdoutpipe [1]);
    int fd = stdoutpipe [0];
    char buf [4];
    int starting = 1;  /* print the prefix when we first read a char */
    while (1) {
      int r = read (fd, buf, 1);
      if (r <= 0) {
        if (r < 0) perror ("print_prefix/read");
        exit (0);
      }
      if (starting) {
        prt_str (1, prefix);
        starting = 0;
      }
      write (1, buf, 1);
      if (buf [0] == '\n')
        starting = 1;   /* print the prefix at the start of the next line */
    }
  } else {   /* parent, replace stdout with the write end of the pipe */
    dup2 (stdoutpipe [1], replace_fd);
    close (stdoutpipe [0]);
    close (stdoutpipe [1]);
  }
}

static void child (int * pipes)
{
  close (pipes [1]);
  close (pipes [2]);
  int read_pipe = pipes [0];
  int write_pipe = pipes [3];

  /* 5 parameters: the program name, two pipes, and the unix socket,
   * plus one for null */
  char * argv [4];
  argv [0] = "acache";
  argv [1] = itoa (read_pipe);
  argv [2] = itoa (write_pipe);
  argv [3] = NULL;

  printf ("in child\n");

  print_prefix ("acache: ", 1);
  print_prefix ("(stderr): ", 2);

  execv ("../acache", argv);
  printf ("error: child returned\n");
}

int main (int argc, char ** argv)
{
  /* create two pipes that we pretend are from ad */
#define NUM_PIPES	2
  int pipes [NUM_PIPES * 2];
  init_pipes (pipes, NUM_PIPES);

  if (fork ())
    parent (pipes);
  else
    child (pipes);
  return 0;
}
