/* tad.c: test ad.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "../packet.h"
#include "../pipemsg.h"
#include "../priority.h"

static void print_buffer (const char * buffer, int count)
{
  int i;
  printf ("%d bytes:", count);
  for (i = 0; i < count && i < 20; i++)
    printf (" %02x", buffer [i] & 0xff);
  if (i < count)
    printf (" ...");
  printf ("\n");
}

/* copied from tsocial.c */
static void init_dest (char * dest, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    dest [i] = (first + i) & 0xff;
  print_buffer (dest, count);
}

static void init_pipes (int * pipes, int num_pipes)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (pipe (pipes + i * 2) < 0) {
      perror ("pipe");
      printf ("error creating pipe set %d\n", i);
      exit (1);
    }
  }
}

static void init_packet (char * p, int psize, int type,
                         int hops, int max_hops, int nbits, int first)
{
  p [0] = type;
  p [1] = hops;
  p [2] = max_hops;
  p [3] = nbits;
  init_dest (p + 4, first, psize - 4);
}

static void print_which_pipe (int pipe, int * pipes, int count)
{
  int i;
  for (i = 0; i < count; i++) {
    if (pipes [i] == pipe) {
      printf (" ([%d]", i);
      if (i == 0) printf (": alocal)");
      if (i == 1) printf (": acache)");
      if (i == 2) printf (": aip)");
      if (i >= 3) printf (": abc [%d])", i - 3);
    }
  }
}

static void receive_all_messages (int * pipes, int count)
{
  while (1) {
    char * message;
    int pipe, pri;
    int r = receive_pipe_message_any (1000, &message, -1, &pipe, &pri);
    if (r <= 0) {
      printf ("no more messages\n");
      break;
    }
    printf ("received %d bytes on pipe %d", r, pipe);
    print_which_pipe (pipe, pipes, count);
    printf (", priority ");
    print_fraction (pri, NULL);
    printf ("\n");
    print_buffer (message, r);
    free (message);
  }
}

static void parent (int * pipes, int num_pipes)
{
  int i;
  int * read_pipes = malloc (num_pipes / 2 * sizeof (int));
  for (i = 0; i < num_pipes; i += 2) {
    close (pipes [2 * i]);
    close (pipes [2 * i + 3]);
    read_pipes [i / 2] = pipes [2 * i + 2];
  }
#define SEND_FROM_ALOCAL	1   /* send packets as if alocal was sending */
#define RECV_TO_ALOCAL		2   /* receive packets intended for alocal */
#define SEND_FROM_ACACHE	5
#define RECV_TO_ACACHE		6
#define SEND_FROM_AIP		9
#define RECV_TO_AIP		10
#define SEND_FROM_ARCV		13
#define RECV_TO_ABC		14

  printf ("in parent\n");
  add_pipes (read_pipes, num_pipes / 2);
  sleep (1);

  /* send a packet from alocal, make sure it is delivered to all 4 dests */
  char packet [200];
  int now = time (NULL);
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 3, 7, 16, now); 
  printf ("sending data packet from local, 3 -> 7 hops, 7/8 priority\n");
  send_pipe_message (pipes [SEND_FROM_ALOCAL], packet, sizeof (packet),
                     SEVEN_EIGHTS);
  receive_all_messages (read_pipes, num_pipes / 2);
  sleep (2);

  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA_ACK,
               0, 1, 8, now + 100); 
  printf ("sending ack packet from aip, 0 -> 1 hops, 0 priority\n");
  send_pipe_message (pipes [SEND_FROM_AIP], packet, sizeof (packet), 0);
  receive_all_messages (read_pipes, num_pipes / 2);

  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA_REQ,
               0, 1, 8, now + 120); 
  printf ("sending req packet from acache, 0 -> 1 hops, 1/2 priority\n");
  send_pipe_message (pipes [SEND_FROM_ACACHE], packet, sizeof (packet),
                     ONE_HALF);
  receive_all_messages (read_pipes, num_pipes / 2);
  sleep (2);

  init_packet (packet, sizeof (packet), ALLNET_TYPE_KEY_XCHG,
               0, 1, 0, now + 132); 
  printf ("sending key exchange packet from aip, 0 -> 1 hops, 0 priority\n");
  send_pipe_message (pipes [SEND_FROM_AIP], packet, sizeof (packet), 0);
  receive_all_messages (read_pipes, num_pipes / 2);

  init_packet (packet, sizeof (packet), ALLNET_TYPE_KEY_XCHG,
               0, 1, 0, now + 154); 
  printf ("  and the response from alocal, 0 -> 1 hops, priority 1\n");
  send_pipe_message (pipes [SEND_FROM_ALOCAL], packet, sizeof (packet),
                     MAX_PRIORITY);
  receive_all_messages (read_pipes, num_pipes / 2);

  sleep (10);
}

static char * itoa (int n)
{
  char * result = malloc (12);  /* plenty of bytes, easier than being exact */
  snprintf (result, 10, "%d", n);
  return result;
}

static void prt_str (int fd, const char * str)
{
  while (*str != '\0')
    write (fd, str++, 1);
}

static void print_prefix (const char * prefix)
{
  int stdoutpipe [2];
  pipe (stdoutpipe);
  if (fork () == 0) {   /* child, print prefix */
    close (stdoutpipe [1]);
    int fd = stdoutpipe [0];
    char buf [4];
    int starting = 1;  /* print the prefix when we first read a char */
    while (1) {
      int r = read (fd, buf, 1);
      if (r <= 0) {
        if (r < 0) perror ("print_prefix/read");
        exit (0);
      }
      if (starting) {
        prt_str (1, prefix);
        starting = 0;
      }
      write (1, buf, 1);
      if (buf [0] == '\n')
        starting = 1;   /* print the prefix at the start of the next line */
    }
  } else {   /* parent, replace stdout with the write end of the pipe */
    dup2 (stdoutpipe [1], 1);
    close (stdoutpipe [0]);
    close (stdoutpipe [1]);
  }
}

static void child (int * pipes, int num_pipes)
{
  int i;
  for (i = 0; i < num_pipes; i += 2) {
    close (pipes [2 * i + 1]);
    close (pipes [2 * i + 2]);
  }

  /* 2 parameters for the program name and the number of pipes, then n
   * parameters for the pipes themselves, and one for NULL */
  char ** argv = malloc ((num_pipes + 3) * sizeof (char *));
  argv [0] = "ad";
  argv [1] = itoa (4);
  for (i = 0; i < num_pipes; i += 2) {
    argv [i + 2] = itoa (pipes [2 * i]);
    argv [i + 3] = itoa (pipes [2 * i + 3]);
  }
  argv [num_pipes + 2] = NULL;

  printf ("in child\n");

  print_prefix ("ad: ");

  execv ("../ad", argv);
  printf ("error: child returned\n");
}

int main (int argc, char ** argv)
{
  /* create two pipes each for alocal, acache, aip, and one abc/arcv */
#define NUM_PIPES	8
  int pipes [NUM_PIPES * 2];
  init_pipes (pipes, NUM_PIPES);

  if (fork ())
    parent (pipes, NUM_PIPES);
  else
    child (pipes, NUM_PIPES);
  return 0;
}
