/* tip.c: test aip.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <fcntl.h>

#include "../packet.h"
#include "../pipemsg.h"
#include "../priority.h"
#include "../util.h"

/* copied from tsocial.c */
static void init_array (char * data, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    data [i] = (first + i) & 0xff;
  print_buffer (data, count, "init", 18, 1);
}

static void init_pipes (int * pipes, int num_pipes)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (pipe (pipes + i * 2) < 0) {
      perror ("pipe");
      printf ("error creating pipe set %d\n", i);
      exit (1);
    }
  }
}

static void init_packet (char * p, int psize, int type,
                         int hops, int max_hops, int nbits, int first)
{
  p [0] = type;
  p [1] = hops;
  p [2] = max_hops;
  p [3] = nbits;
  init_array (p + 4, first, psize - 4);
}

static void send_mapping (char * socket_name, char * dest, int nbits,
                          struct sockaddr * addr)
{
  int sock = socket (AF_UNIX, SOCK_DGRAM, 0);
  socklen_t sl = sizeof (struct sockaddr_un);
  struct sockaddr_un sun;
  sun.sun_family = AF_UNIX;
  strcpy (sun.sun_path, socket_name);
  int bsize = 32;
  char buffer [32];
  bzero (buffer, sizeof (buffer));
  memcpy (buffer, dest, DESTINATION_SIZE);
  buffer [DESTINATION_SIZE    ] = nbits;
  int port = 0;
  if (addr->sa_family == AF_INET) {
    struct sockaddr_in * sin = (struct sockaddr_in *) addr;
    buffer [DESTINATION_SIZE + 1] = 4;
    buffer [DESTINATION_SIZE + 2] = ((sin->sin_port) >> 8) & 0xff;
    buffer [DESTINATION_SIZE + 3] = ((sin->sin_port)     ) & 0xff;
    memcpy (buffer + DESTINATION_SIZE + 4, (char *) &(sin->sin_addr), 4);
  } else {
    struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) addr;
    buffer [DESTINATION_SIZE + 1] = 6;
    buffer [DESTINATION_SIZE + 2] = ((sin6->sin6_port) >> 8) & 0xff;
    buffer [DESTINATION_SIZE + 3] = ((sin6->sin6_port)     ) & 0xff;
    memcpy (buffer + DESTINATION_SIZE + 4, (char *) &(sin6->sin6_addr), 16);
  }
  print_buffer (buffer, bsize, "sending map", 32, 1);
  int result = sendto (sock, buffer, bsize, 0, (struct sockaddr *) (&sun), sl);
  if (result < bsize) {
    perror ("sendto");
    printf ("unable to send %d bytes to ", bsize);
    print_sockaddr ((struct sockaddr *) (&sun), 0);
    printf (", result is %d\n", result);
  }
  close (sock);
}

static void send_udp_packet (int sock, char * packet, int psize,
                             struct sockaddr * s)
{
  int opened = 0;
  if (sock == -1) {
    sock = socket (s->sa_family, SOCK_DGRAM, 0);
    opened = 1;
  }
  socklen_t sl = sizeof (struct sockaddr_in6);
  if (s->sa_family == AF_INET) sl = sizeof (struct sockaddr_in);
  int result = sendto (sock, packet, psize, 0, s, sl);
  if (result < psize) {
    perror ("sendto");
    printf ("unable to send %d bytes to ", psize);
    print_sockaddr (s, 0);
    printf (", result is %d\n", result);
  }
  if (opened)
    close (sock);
}

static void receive_all_messages (int * pipes, int count, int udp)
{
  while (1) {
    char * message;
    int pipe, pri;
    int r = receive_pipe_message_any (1000, &message, udp, &pipe, &pri);
    if ((r == 0) && (pipe == udp)) {
      message = malloc_or_fail (ALLNET_MTU, "tip message");
      struct sockaddr sa;
      socklen_t sl = sizeof (sa);
      r = recvfrom (udp, message, ALLNET_MTU, MSG_DONTWAIT, &sa, &sl);
      printf ("received %d udp bytes from ", r);
      print_sockaddr (&sa, sl);
      printf ("\n");
    }
    if (r <= 0) {
      printf ("no more messages\n");
      break;
    }
    printf ("received %d bytes on pipe %d", r, pipe);
    /* print_which_pipe (pipe, pipes, count); */
    printf (", priority ");
    print_fraction (pri, NULL);
    printf ("\n");
    print_buffer (message, r, NULL, 20, 1);
    free (message);
  }
}

static int new_local_socket (int ipv6)
{
  int fd = socket (ipv6 ? AF_INET6 : AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
    perror ("socket");
    exit (1);
  }
  struct sockaddr_storage address;
  struct sockaddr     * ap  = (struct sockaddr     *) &address;
  struct sockaddr_in  * ap4 = (struct sockaddr_in  *) ap;
  struct sockaddr_in6 * ap6 = (struct sockaddr_in6 *) ap;
  socklen_t addr_size = sizeof (address);

  if (ipv6) {
    ap6->sin6_family = PF_INET6;
    memcpy (&(ap6->sin6_addr), &(in6addr_loopback), sizeof (ap6->sin6_addr));
    ap6->sin6_port = ALLNET_PORT;
  } else {
    ap4->sin_family = AF_INET;
    ap4->sin_addr.s_addr = htonl (INADDR_LOOPBACK);
    ap4->sin_port = ALLNET_PORT;
  }
  
  if (connect (fd, ap, addr_size) < 0) {
    perror ("connect");
    exit (1);
  }
  return fd;
}

static void parent (int * pipes, char * unix_socket_name)
{
  close (pipes [0]);
  close (pipes [3]);
  int write_pipe = pipes [1];
  int read_pipe = pipes [2];

  int read_pipes [5];
  read_pipes [0] = read_pipe;

  printf ("in parent\n");
  add_pipes (read_pipes, 1);
  sleep (1);

  int udp = socket (AF_INET6, SOCK_DGRAM, 0);
  if (udp < 0) {
    perror ("main loop socket");
    exit (1);
  }
  printf ("opened UDP socket %d\n", udp);

  char packet [200];
  int now = time (NULL);
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 0, 1, 8, now); 

  /* send some packets from different udp ports, should send back to those */
  struct sockaddr_in6 sa6;
  sa6.sin6_family = AF_INET6;
  memcpy (&(sa6.sin6_addr), &(in6addr_loopback), sizeof (sa6.sin6_addr));
  sa6.sin6_port = ALLNET_PORT;
  printf ("sending 4 UDP data packets, 0 -> 1 hops\n");
  send_udp_packet ( -1, packet, 100, (struct sockaddr *) (&sa6));
  send_udp_packet ( -1, packet,  95, (struct sockaddr *) (&sa6));
  send_udp_packet ( -1, packet,  90, (struct sockaddr *) (&sa6));
  send_udp_packet (udp, packet, 128, (struct sockaddr *) (&sa6));
  sleep (2);

  /* send a packet from ad, nothing should really happen... */
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 3, 7, 16, now + 78); 
  printf ("sending data packet from ad, 3 -> 7 hops, 1/4 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet), ONE_QUARTER);
  receive_all_messages (read_pipes, 1, udp);
  sleep (2);

  /* open three sockets and repeat, the message should be deliverd */
  read_pipes [1] = new_local_socket (0);
  read_pipes [2] = new_local_socket (1);
  read_pipes [3] = new_local_socket (1);
  add_pipes (read_pipes + 1, 3);
  sleep (2);
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 0, 5, 16, now + 31); 
  printf ("sending data packet from remote, 0 -> 5 hops, 1/2 priority\n");
  send_pipe_message (read_pipes [2], packet, sizeof (packet) - 10, ONE_HALF);
  receive_all_messages (read_pipes, 4, udp);
  sleep (1);

  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 1, 1, 16, now + 100); 
  printf ("sending data packet from ad, 1 -> 1 hops, 1/8 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet) - 10, ONE_EIGHT);
  receive_all_messages (read_pipes, 4, udp);
  sleep (1);

  /* closing one "client" socket should close the socket for aip as well */
  close (read_pipes [1]);
  remove_pipe (read_pipes [1]);
  read_pipes [1] = read_pipes [3];
  sleep (2);

  /* the other socket should still work */
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 9, 9, 16, now + 200); 
  printf ("sending data packet from ad, 9 -> 9 hops, 1/4 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet), ONE_QUARTER);
  receive_all_messages (read_pipes, 3, udp);
  sleep (2);

  /* send a mapping for the address in this packet */
  struct sockaddr_storage ssa;
  struct sockaddr * sa = (struct sockaddr *) (&ssa);
  socklen_t sl = sizeof (ssa);
  if (getsockname (udp, sa, &sl) == 0) {
    send_mapping (unix_socket_name, packet + 4, 8, sa);
    sleep (2);
  /* resend (almost) the same packet after sending a mapping for its address */
    (packet [10])--;
    send_pipe_message (write_pipe, packet, 48, THREE_QUARTERS);
    receive_all_messages (read_pipes, 3, udp);
    sleep (2);
  }

  printf ("tip main thread exiting\n");
}

static char * itoa (int n)
{
  char * result = malloc (12);  /* plenty of bytes, easier than being exact */
  snprintf (result, 10, "%d", n);
  return result;
}

static void prt_str (int fd, const char * str)
{
  while (*str != '\0')
    write (fd, str++, 1);
}

static void print_prefix (const char * prefix, int replace_fd)
{
  /* for redirecting the output to a terminal, do "who am i" in that terminal,
     and put the terminal name here */
/*
  int tty_fd = open ("/dev/pts/3", O_WRONLY);
  dup2 (tty_fd, 1);
  return;
*/
  int stdoutpipe [2];
  pipe (stdoutpipe);
  if (fork () == 0) {   /* child, print prefix */
    close (stdoutpipe [1]);
    int fd = stdoutpipe [0];
    char buf [4];
    int starting = 1;  /* print the prefix when we first read a char */
    while (1) {
      int r = read (fd, buf, 1);
      if (r <= 0) {
        if (r < 0) perror ("print_prefix/read");
        exit (0);
      }
      if (starting) {
        prt_str (1, prefix);
        starting = 0;
      }
      write (1, buf, 1);
      if (buf [0] == '\n')
        starting = 1;   /* print the prefix at the start of the next line */
    }
  } else {   /* parent, replace stdout with the write end of the pipe */
    dup2 (stdoutpipe [1], replace_fd);
    close (stdoutpipe [0]);
    close (stdoutpipe [1]);
  }
}

static void child (int * pipes, char * unix_socket)
{
  close (pipes [1]);
  close (pipes [2]);
  int read_pipe = pipes [0];
  int write_pipe = pipes [3];

  /* 5 parameters: the program name, two pipes, and the unix socket,
   * plus one for null */
  char * argv [5];
  argv [0] = "aip";
  argv [1] = itoa (read_pipe);
  argv [2] = itoa (write_pipe);
  argv [3] = unix_socket;
  argv [4] = NULL;

  printf ("in child\n");

  print_prefix ("aip: ", 1);
  print_prefix ("(stderr): ", 2);

  execv ("../aip", argv);
  printf ("error: child returned\n");
}

int main (int argc, char ** argv)
{
  /* create two pipes that we pretend are from ad */
#define NUM_PIPES	2
  int pipes [NUM_PIPES * 2];
  init_pipes (pipes, NUM_PIPES);

  char * unix_socket_name = "/tmp/allnet_addrs";

  if (fork ())
    parent (pipes, unix_socket_name);
  else
    child (pipes, unix_socket_name);
  printf ("removing %s\n", unix_socket_name);
  if (unlink (unix_socket_name) < 0)
    perror ("unlink");
  return 0;
}
