/* tlocal.c: test alocal.c */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>

#include "../packet.h"
#include "../pipemsg.h"
#include "../priority.h"

static void print_buffer (const char * buffer, int count)
{
  int i;
  printf ("%d bytes:", count);
  for (i = 0; i < count && i < 18; i++)
    printf (" %02x", buffer [i] & 0xff);
  if (i < count)
    printf (" ...");
  printf ("\n");
}

/* copied from tsocial.c */
static void init_array (char * data, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    data [i] = (first + i) & 0xff;
  printf ("init ");
  print_buffer (data, count);
}

static void init_pipes (int * pipes, int num_pipes)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (pipe (pipes + i * 2) < 0) {
      perror ("pipe");
      printf ("error creating pipe set %d\n", i);
      exit (1);
    }
  }
}

static void init_packet (char * p, int psize, int type,
                         int hops, int max_hops, int nbits, int first)
{
  p [0] = type;
  p [1] = hops;
  p [2] = max_hops;
  p [3] = nbits;
  init_array (p + 4, first, psize - 4);
}

static void receive_all_messages (int * pipes, int count)
{
  while (1) {
    char * message;
    int pipe, pri;
    int r = receive_pipe_message_any (1000, &message, -1, &pipe, &pri);
    if (r <= 0) {
      printf ("no more messages\n");
      break;
    }
    printf ("received %d bytes on pipe %d", r, pipe);
    /* print_which_pipe (pipe, pipes, count); */
    printf (", priority ");
    print_fraction (pri, NULL);
    printf ("\n");
    print_buffer (message, r);
    free (message);
  }
}

static int new_local_socket (int ipv6)
{
  int fd = socket (ipv6 ? AF_INET6 : AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
    perror ("socket");
    exit (1);
  }
  struct sockaddr_storage address;
  struct sockaddr     * ap  = (struct sockaddr     *) &address;
  struct sockaddr_in  * ap4 = (struct sockaddr_in  *) ap;
  struct sockaddr_in6 * ap6 = (struct sockaddr_in6 *) ap;
  socklen_t addr_size = sizeof (address);

  if (ipv6) {
    ap6->sin6_family = PF_INET6;
    memcpy (&(ap6->sin6_addr), &(in6addr_loopback), sizeof (ap6->sin6_addr));
    ap6->sin6_port = ALLNET_LOCAL_PORT;
  } else {
    ap4->sin_family = AF_INET;
    ap4->sin_addr.s_addr = htonl (INADDR_LOOPBACK);
    ap4->sin_port = ALLNET_LOCAL_PORT;
  }
  
  if (connect (fd, ap, addr_size) < 0) {
    perror ("connect");
    exit (1);
  }
  return fd;
}

static void parent (int * pipes)
{
  close (pipes [0]);
  close (pipes [3]);
  int write_pipe = pipes [1];
  int read_pipe = pipes [2];

  int read_pipes [4];
  read_pipes [0] = read_pipe;

  printf ("in parent\n");
  add_pipes (read_pipes, 1);
  sleep (1);

  /* send a packet from ad, nothing should really happen... */
  char packet [200];
  int now = time (NULL);
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 3, 7, 16, now); 
  printf ("sending data packet from ad, 3 -> 7 hops, 1/4 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet), ONE_QUARTER);
  receive_all_messages (read_pipes, 1);
  sleep (2);

  /* open two sockets and repeat, the message should be deliverd */
  read_pipes [1] = new_local_socket (0);
  read_pipes [2] = new_local_socket (1);
  add_pipes (read_pipes + 1, 2);
  sleep (2);
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 0, 5, 16, now + 31); 
  printf ("sending data packet from local, 0 -> 5 hops, 1/2 priority\n");
  send_pipe_message (read_pipes [2], packet, sizeof (packet) - 10, ONE_HALF);
  receive_all_messages (read_pipes, 3);
  sleep (3);

  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 1, 1, 16, now + 100); 
  printf ("sending data packet from ad, 1 -> 1 hops, 1/8 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet) - 10, ONE_EIGHT);
  receive_all_messages (read_pipes, 3);
  sleep (10);

  /* closing our "client" socket should close the socket for alocal as well */
  close (read_pipes [1]);
  remove_pipe (read_pipes [1]);
  read_pipes [1] = read_pipes [2];
  sleep (2);

  /* the other pipe should still work */
  init_packet (packet, sizeof (packet), ALLNET_TYPE_DATA, 9, 9, 16, now + 200); 
  printf ("sending data packet from ad, 9 -> 9 hops, 1/4 priority\n");
  send_pipe_message (write_pipe, packet, sizeof (packet), ONE_QUARTER);
  receive_all_messages (read_pipes, 2);
  sleep (2);

  sleep (10);
  printf ("tlocal main thread exiting\n");
}

static char * itoa (int n)
{
  char * result = malloc (12);  /* plenty of bytes, easier than being exact */
  snprintf (result, 10, "%d", n);
  return result;
}

static void prt_str (int fd, const char * str)
{
  while (*str != '\0')
    write (fd, str++, 1);
}

static void print_prefix (const char * prefix, int replace_fd)
{
  /* for redirecting the output to a terminal, do "who am i" in that terminal,
     and put the terminal name here */
/*
  int tty_fd = open ("/dev/pts/0", O_WRONLY);
  dup2 (tty_fd, 1);
  return;
*/
  int stdoutpipe [2];
  pipe (stdoutpipe);
  if (fork () == 0) {   /* child, print prefix */
    close (stdoutpipe [1]);
    int fd = stdoutpipe [0];
    char buf [4];
    int starting = 1;  /* print the prefix when we first read a char */
    while (1) {
      int r = read (fd, buf, 1);
      if (r <= 0) {
        if (r < 0) perror ("print_prefix/read");
        exit (0);
      }
      if (starting) {
        prt_str (1, prefix);
        starting = 0;
      }
      write (1, buf, 1);
      if (buf [0] == '\n')
        starting = 1;   /* print the prefix at the start of the next line */
    }
  } else {   /* parent, replace stdout with the write end of the pipe */
    dup2 (stdoutpipe [1], replace_fd);
    close (stdoutpipe [0]);
    close (stdoutpipe [1]);
  }
}

static void child (int * pipes)
{
  close (pipes [1]);
  close (pipes [2]);
  int read_pipe = pipes [0];
  int write_pipe = pipes [3];

  /* 4 parameters: the program name and two pipes, plus one for null */
  char * argv [4];
  argv [0] = "alocal";
  argv [1] = itoa (read_pipe);
  argv [2] = itoa (write_pipe);
  argv [3] = NULL;

  printf ("in child\n");

  print_prefix ("alocal: ", 1);
  print_prefix ("(stderr): ", 2);

  execv ("../alocal", argv);
  printf ("error: child returned\n");
}

int main (int argc, char ** argv)
{
  /* create two pipes that we pretend are from ad */
#define NUM_PIPES	2
  int pipes [NUM_PIPES * 2];
  init_pipes (pipes, NUM_PIPES);

  if (fork ())
    parent (pipes);
  else
    child (pipes);
  return 0;
}
