/* tping.c: test abc.c and arcv.c: send a packet every sec, print responses */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../packet.h"
#include "../pipemsg.h"
#include "../priority.h"

static void print_buffer (const char * buffer, int count)
{
  int i;
  printf ("%d bytes:", count);
  for (i = 0; i < count && i < 20; i++)
    printf (" %02x", buffer [i] & 0xff);
  if (i < count)
    printf (" ...");
  printf ("\n");
}

/* copied from tsocial.c */
static void init_array (char * data, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    data [i] = (first + i) & 0xff;
/*  print_buffer (data, count); */
}

static void init_pipes (int * pipes, int num_pipes)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (pipe (pipes + i * 2) < 0) {
      perror ("pipe");
      printf ("error creating pipe set %d\n", i);
      exit (1);
    }
  }
}

static void init_packet (char * p, int psize, int type, int hops, int max_hops,
                         int nbits, char * dest, int first)
{
  p [0] = type;
  p [1] = hops;
  p [2] = max_hops;
  p [3] = nbits;
  memset (p + 4, 0, DESTINATION_SIZE);
  int nbytes = (nbits + 7) / 8;
  memcpy (p + 4, dest, nbytes);
  int offset = 4 + DESTINATION_SIZE;
  init_array (p + offset, first, psize - offset);
}

static int receive_all_messages (int * pipes, int count)
{
  int received = 0;   /* how many messages we have received */
  while (1) {
    char * message;
    int pipe, pri;
    int r = receive_pipe_message_any (1000, &message, -1, &pipe, &pri);
    if (r <= 0) {
      /* printf ("no more messages\n"); */
      break;
    }
    printf ("received %d bytes on pipe %d from abc", r, pipe);
    printf (", priority ");
    print_fraction (pri, NULL);
    if ((r > 4) && (message [0] != ALLNET_TYPE_MGMT))
      printf (", packet number %d\n", message [4] & 0xff);
    else
      printf ("\n");
    if (r >= ALLNET_BASIC_HEADER_SIZE) {
      printf ("type ");
      switch (message [0] & 0xff) {
      case ALLNET_TYPE_DATA: printf ("data"); break;
      case ALLNET_TYPE_DATA_ACK: printf ("ack"); break;
      case ALLNET_TYPE_DATA_REQ: printf ("req"); break;
      case ALLNET_TYPE_KEY_XCHG: printf ("key"); break;
      case ALLNET_TYPE_CLEAR: printf ("clear"); break;
      case ALLNET_TYPE_MGMT: printf ("management"); break;
      default:
        printf ("unknown %d/0x%x", message [0] & 0xff, message [0] & 0xff);
        break;
      }
      printf (", hops: %d -> %d, ", message [1] & 0xff, message [2] & 0xff);
      printf ("%d bits of address: ", message [3] & 0xff);
      int bytes = ((message [3] & 0xff) + 7) / 8;
      if (bytes > DESTINATION_SIZE) bytes = DESTINATION_SIZE;
      if (bytes > 0)                print_buffer (message + 4, bytes);
      else                          printf ("\n");
    }
    print_buffer (message, r);
    free (message);
    received++;
  }
  return received;
}

/* argument: pipe from arcv, pipe to abc */
static void parent (int * pipes)
{
  printf ("in parent\n");

  close (pipes [1]);
  close (pipes [2]);
  int abc_read_pipe = pipes [0];
  int abc_write_pipe = pipes [3];
  add_pipes (&abc_read_pipe, 1);

  int count = 0;
  int received = 0;
  while (1) {
    char packet [200];
    char dest [DESTINATION_SIZE];
    int type = ALLNET_TYPE_CLEAR;
    bzero (dest, DESTINATION_SIZE);
    init_array (dest, count, 2);
    init_packet (packet, sizeof (packet), type, 0, 1, 16, dest, count); 
    printf ("sending data packet %d from ad, 0 -> 1 hops, 7/8 priority\n",
            count);
    if (! send_pipe_message (abc_write_pipe, packet, sizeof (packet),
                             SEVEN_EIGHTS))
      break;
    received += receive_all_messages (&abc_read_pipe, 1);
    sleep (1);
    count++;
    if ((count >= 250) || (received >= 5))
      return;
  }
  sleep (10);
  printf ("tping exiting\n");
}

static char * itoa (int n)
{
  char * result = malloc (12);  /* plenty of bytes, easier than being exact */
  snprintf (result, 10, "%d", n);
  return result;
}

static void prt_str (int fd, const char * str)
{
  while (*str != '\0')
    write (fd, str++, 1);
}

static void print_prefix (const char * prefix, int replace_fd)
{
  int stdoutpipe [2];
  pipe (stdoutpipe);
  if (fork () == 0) {   /* child, print prefix */
    close (stdoutpipe [1]);
    int fd = stdoutpipe [0];
    char buf [4];
    int starting = 1;  /* print the prefix when we first read a char */
    while (1) {
      int r = read (fd, buf, 1);
      if (r <= 0) {
        if (r < 0) perror ("print_prefix/read");
        exit (0);
      }
      if (starting) {
        prt_str (1, prefix);
        starting = 0;
      }
      write (1, buf, 1);
      if (buf [0] == '\n')
        starting = 1;   /* print the prefix at the start of the next line */
    }
  } else {   /* parent, replace stdout with the write end of the pipe */
    dup2 (stdoutpipe [1], replace_fd);
    close (stdoutpipe [0]);
    close (stdoutpipe [1]);
  }
}

static void call_abc (int * pipes, char * interface)
{
  close (pipes [0]);
  close (pipes [3]);
  int write_pipe = pipes [1];
  int read_pipe = pipes [2];

  /* 4 parameters: the program name, two pipes pipe, and an interface,
                   plus one array entry for null */
  char * argv [5];
  argv [0] = "abc";
  argv [1] = itoa (read_pipe);
  argv [2] = itoa (write_pipe);
  argv [3] = interface;
  argv [4] = NULL;

  printf ("in child: starting abc\n");

  print_prefix ("abc: ", 1);
  print_prefix ("(stderr): ", 2);

  execv ("../abc", argv);
  printf ("error: abc child returned\n");
}

int main (int argc, char ** argv)
{
  /* create two pipes for abc */
#define NUM_PIPES	2
  int pipes [NUM_PIPES * 2];
  init_pipes (pipes, NUM_PIPES);
  char interface [] = "wlan0";

  if (fork ()) {
    parent (pipes);
  } else {
    call_abc (pipes, interface);
  }
  return 0;
}
