
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>

#include "../packet.h"
#include "../social.h"

static void init_dest (char * dest, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    dest [i] = (first + i) & 0xff;
  printf ("init %d bytes:", count);
  for (i = 0; i < count && i < 20; i++)
    printf (" %02x", dest [i] & 0xff);
  if (i < count)
    printf (" ...");
  printf ("\n");
}

static void write_dest (char * dest, char * fname, int count)
{
  int fd = open (fname, O_WRONLY | O_APPEND | O_CREAT, 0600);
  if (fd < 0) {
    perror ("open");
    printf ("unable to open file %s for writing\n", fname);
    exit (1);
  }
  static char buf [1000000];
  if (sizeof (buf) < count * 2 + 2) {
    printf ("using static buf of size %zd, but needed %d (from %d)\n",
            sizeof (buf), count * 2 + 2, count);
    exit (1);
  }
  int i;
  for (i = 0; i < count; i++)
    snprintf (buf + i * 2, sizeof (buf) - (i * 2), "%02x", dest [i] & 0xff);
  snprintf (buf + i * 2, 2, "\n");
  int len = strlen (buf);
  int w = write (fd, buf, len);
  if (w != len) {
    perror ("write");
    printf ("unable to write %d bytes to file %s, wrote %d\n", len, fname, w);
    exit (1);
  }
  close (fd);
  if (strlen (buf) > 50)
    snprintf (buf + 50, sizeof (buf) - 50, "...\n");
  printf ("wrote %s", buf);
}

int main (int argc, char ** argv)
{
  struct social_info * soc = init_social (10000);

  time_t now = time (NULL);
  time_t next = update_social (soc, 10);
  printf ("now is %ld, next is %ld, diff %ld should be 10\n",
          now, next, next - now);

  /* a non-random number that is different on each run, so we can add
   * different entries to the table */
  int notrandom = time (NULL);

  char dest1 [512];
  char dest2 [16];
  char dest3 [8];
  init_dest (dest1, 0 + notrandom, 512);
  init_dest (dest2, 1 + notrandom, 16);
  init_dest (dest3, 2 + notrandom, 8);

  int conn = social_connection (soc, dest1, 0);
  printf ("tier of a destination with 0 bits is %d\n", conn);

  conn = social_connection (soc, dest1, DESTINATION_SIZE * 8);
  printf ("tier of a destination with 96 bits is %d\n", conn);

  /* create the social connection files with predictable IDs */
  char fname [] = "/home/esb/.allnet/ad/social0";
  char * last = rindex (fname, '0');
  write_dest (dest1, fname, 512);
  *last = '1';
  write_dest (dest2, fname, 16);
  *last = '2';
  write_dest (dest3, fname, 8);

  /* check to see that before update_social, those IDs are not loaded */
  conn = social_connection (soc, dest1, DESTINATION_SIZE * 8);
  printf ("before update_social, tier of destination 0 is %d\n", conn);
  conn = social_connection (soc, dest2, DESTINATION_SIZE * 8);
  printf ("before update_social, tier of destination 1 is %d\n", conn);
  conn = social_connection (soc, dest3, 8 * 8);
  printf ("before update_social, tier of destination 2 is %d\n", conn);

  /* update social, and watch the tiers change */
  now = time (NULL);
  next = update_social (soc, 100);
  printf ("now is %ld, next is %ld, diff %ld should be 100\n",
          now, next, next - now);
  
  /* check to see that those IDs are now loaded */
  conn = social_connection (soc, dest1, DESTINATION_SIZE * 8);
  printf ("after update_social, tier of destination 0 is %d\n", conn);
  conn = social_connection (soc, dest2, DESTINATION_SIZE * 8);
  printf ("after update_social, tier of destination 1 is %d\n", conn);
  conn = social_connection (soc, dest3, 8 * 8);
  printf ("after update_social, tier of destination 2 is %d\n", conn);

  conn = social_connection (soc, dest1, 9);
  printf ("9 bits of destination 0 puts it in tier %d\n", conn);
  conn = social_connection (soc, dest1, 15);
  printf ("15 bits of destination 0 puts it in tier %d\n", conn);
  conn = social_connection (soc, dest1, 20);
  printf ("20 bits of destination 0 puts it in tier %d\n", conn);
  conn = social_connection (soc, dest1, 22);
  printf ("22 bits of destination 0 puts it in tier %d\n", conn);

  return 0;
}

