/* util.h: useful functions used by different programs */

#ifndef ALLNET_UTIL_H
#define ALLNET_UTIL_H

#include <sys/time.h>
#include <sys/socket.h>

/* print up to max of the count characters in the buffer.
 * desc is printed first unless it is null
 * a newline is printed after if print_eol
 */
extern void print_buffer (const char * buffer, int count, char * desc,
                          int max, int print_eol);

/* print an arbitrary socket address */
extern void print_sockaddr (struct sockaddr * sap, int addr_size);

/* return 1 if the first nbits of x match the first nbits of y, else 0 */
/* where nbits is the lesser of xbits and ybits */
extern int matches (unsigned char * x, int xbits, unsigned char * y, int ybits);

/* useful time functions and constants */
#define US_PER_S        1000000    /* microseconds in a second */
#define ONE_SECOND      US_PER_S
#define HALF_SECOND     (US_PER_S / 2)

extern unsigned long long delta_us (struct timeval * t1, struct timeval * t2);

extern void add_us (struct timeval * t, unsigned long long us);

/* set result to a random time between start + min and start + max */
extern void set_time_random (struct timeval * start, unsigned long long min,
                             unsigned long long max, struct timeval * result);

/* if malloc is not successful, exit after printing */
extern void * malloc_or_fail (int bytes, char * desc);

/* place the values 0..n-1 at random within the given array */
extern void random_permute_array (int n, int * array);

/* malloc and return an n-element int array containing the values 0..n-1
 * in some random permuted order */
extern int * random_permute (int n);

#endif /* ALLNET_UTIL_H */
